
crossValidate <- function(formula, data, ...) {
  mc <- match.call(expand.dots = TRUE)
  m <- match(c("formula", "data", "id", "family", "weights",
               "vsplits", "userseed", "nfolds", "silent"), names(mc), 0)
  mc <- mc[c(1, m)]
  mc[[1]] <- as.name("DSA")

  ## get the correct size using the design matrix. 
  mm <- model.matrix(terms(formula, data = data), data = data)[, -1, drop=FALSE]
  mc$maxsize <- size <- length(colnames(mm))

  mc$maxsumofpow <- mc$maxorderint <- 1
  mc$cross.validate.model.selected <- FALSE
  mc$expand.forced.term.factors <- TRUE
  res <- eval(mc, parent.frame())
  return(res$average.CVrisks[1, size + 1])
}

DSA <- function(formula, data, id = 1:nrow(data), family = gaussian, weights = NULL,
                maxsize, maxorderint, maxsumofpow, vsplits = NULL,
                userseed = NULL, nfolds = 5, model.matrices = FALSE, silent = TRUE, ...)
{
  extra.args <- list(...)
  start.time <- as.POSIXlt(Sys.time())
  call <- match.call()
  model.mats <- get.model.matrices(formula, data)
  
  if (!missing(silent)) {
    ## get the old one so we can reset it when we leave.
    old.message.level <- getDSAMessageLevel()
    if (silent)
      setDSAMessageLevel(-1)
    else
      setDSAMessageLevel(0)
  }

  ## in this case, we have a model where the factors and the data frame line up;
  ## ie. a call from the DSA to itself in order to cross-validate the final model.
  if (is.null(expand.forced.term.factors <- extra.args$expand.forced.term.factors))
    expand.forced.term.factors <- TRUE
  
  ## in this case, either the DSA is calling itself or someone called the crossValidate
  ## method.
  if (is.null(cross.validate.model.selected <- extra.args$cross.validate.model.selected))
    cross.validate.model.selected <- TRUE

  if (!is.logical(expand.forced.term.factors) ||
      !is.logical(cross.validate.model.selected))
    stop("Error calling DSA, logical parameters must be logical.")
  
  Xlearn <- t(model.mats$X)
  Ylearn <- model.mats$Y
    
  # save the names from the original.
  Ylearn.orig.names <- dimnames(Ylearn)[[2]]

  if (model.matrices == TRUE) {
    return(model.mats)
  }
  
  if(is.null(vsplits)) {
    # compute splits. 
    if (!missing(userseed) & is.numeric(userseed))
      set.seed(userseed)

    uniqueIDlearn <- t(as.matrix(unique(id)))
    nIDlearn <- ncol(uniqueIDlearn)
    
    #each unique ID is assigned to a split:
    splitvector <- c(rep(1:nfolds, each = floor(nIDlearn/nfolds)),
                     rep(nfolds, nIDlearn - floor(nIDlearn/nfolds)*nfolds))
    splitvector <- sample(splitvector)

    #each observation is assigned to the split number to which its
    #corresponding ID was assigned
    splitvectorlearn <- apply(t(id), 2, function(x,y,z) {
      return(y[z==x])
    }, y=splitvector,z=uniqueIDlearn)
    
    vsplits <- apply(t(splitvectorlearn),2,function(x,y) {
      splitassignment <- rep(0,y)
      splitassignment[x] <- 1
      return(splitassignment)
    }, y=nfolds) 
  }
    
  nvarX <- nrow(Xlearn)
  nlearn <- ncol(Xlearn)
  maxnval <- max(apply(vsplits, 1, sum))
  maxntrain <- nlearn - min(apply(vsplits,1,sum))
  minntrain <- nlearn - max(apply(vsplits,1,sum))

  ## construct the forced term matrix - null if no forced terms.
  forcedterms <- get.forced.terms(formula, data, Xlearn, expand.forced.term.factors)
  
  if (!is.null(forcedterms))
    nforced <- nrow(forcedterms)
  else
    nforced <- 0

  ## now check that we have all the parameters correct and we are ready to
  ## sanely enter the C routines.
  if (!is.numeric(id) || (length(id) != ncol(Xlearn)))
    stop("id must be an integer vector with length == nrow(data)")
  if (!(is.numeric(nfolds) && (length(nfolds) == 1) && (nfolds > 1)))
    stop("nfolds must be a positive integer > 1")
  if (is.null(vsplits) && !is.null(userseed)) {
    if (!(is.numeric(userseed) && length(userseed) == 1))
      stop("userseed must be a scalar integer")
  }
  if (!(is.numeric(maxorderint) && maxorderint > 0) ||
      !(is.numeric(maxsumofpow) && maxsumofpow > 0) ||
      !(is.numeric(maxsize) && maxsize >= 0))
    stop("maxorderint, maxsumofpow must be an integer greater than 0")
  
  if (!is.null(weights))
    if (!is.matrix(weights) || !is.numeric(weights))
      stop("weights must be a matrix of real numbers.")
  if (!(is.matrix(vsplits) && is.numeric(vsplits) &&
        all(!is.na(vsplits)) && all(vsplits == 1 || vsplits == 0)))
    stop("vsplits must be a matrix of 0s or 1s.\n")

  ## consistency between the arguments.
  if((length(unique(id))/nfolds) <= 1)
    stop("The number of independent experimental units is too low for nfolds.")
  
  ## allocate space for the C routines.
  CVrisks <- matrix(NA, nrow = maxorderint, ncol = maxsize+1)
  coefficients <- rep(0, maxsize + 1)
  modelselected <- matrix(NA, nrow=maxsize, ncol = nvarX)

  if (is.null(weights)) {
    WTtrainvallearn <- matrix(1, ncol = nlearn, nrow = (nfolds+1))
    useWeights <- 0
  }
  else {
    WTtrainvallearn <- weights
    useWeights <- 1
  }
  
  if (ncol(Ylearn) == 2) {
    binWTtrainvallearn <- Ylearn[,1] + Ylearn[,2]
    
    # now Ylearn becomes the success ratio.
    tempYlearn <- as.matrix(ifelse(binWTtrainvallearn == 0, 0, Ylearn[,1]/binWTtrainvallearn))
    dimnames(tempYlearn)[[2]] <- list(dimnames(Ylearn)[[2]][1])
    Ylearn <- tempYlearn
    
    # and the weights become the number of trials.
    WTtrainvallearn <- t(apply(WTtrainvallearn, 1, function(row) {
      row * binWTtrainvallearn
    }))

    binWTtrainvallearn[is.na(binWTtrainvallearn)] <- 0
    useWeights <- 1
  }
  else {
    binWTtrainvallearn <- rep(1, nlearn)
  }

  ##
  ## this was directly lifted from 'glm'
  if (is.character(family)) 
    family <- get(family, mode = "function", envir = parent.frame())
  if (is.function(family)) 
    family <- family()
  if (is.null(family$family)) {
    print(family)
    stop("'family' not recognized")
  }
  if (family$family == "gaussian")
    binind <- 0
  else if (family$family == "binomial")
    binind <- 1
  
  gCONST <- c(maxsize, maxorderint, maxsumofpow, nfolds, nlearn,
              nvarX, useWeights, NA, NA, binind)

  storage.mode(gCONST) <- "integer"
  storage.mode(Xlearn) <- "double"        
  storage.mode(Ylearn) <- "double"
  storage.mode(CVrisks) <- "double"
  storage.mode(coefficients) <- "double"
  storage.mode(vsplits) <- "integer"
  storage.mode(WTtrainvallearn) <- "double"    
  storage.mode(nforced) <- "integer"
  storage.mode(forcedterms) <- "integer"
  storage.mode(maxnval) <- "integer"    
  storage.mode(maxntrain) <- "integer"    
  storage.mode(modelselected) <- "integer"
  storage.mode(binWTtrainvallearn) <- "double"
  
  res <- .C("DSA_PACK_Rentry",
            gCONST,
            Xlearn,
            Ylearn,
            CVrisks,
            coefficients,
            vsplits,
            WTtrainvallearn,
            nforced,
            forcedterms,
            maxntrain,
            maxnval,
            modelselected,
            binWTtrainvallearn,
            NAOK = TRUE)
  
  gCONST <- res[[1]]
  modelf <- get.final.model(res[[12]], dimnames(Xlearn)[[1]], Ylearn.orig.names)
  coefficients <- res[[5]]
  
  final.time <- as.POSIXlt(Sys.time())
  total.time <- difftime(final.time, start.time)    

  ## now we will cross validate the response. calling the DSA again w/appropriate
  ## flags set. This is similar to 'crossValidate' however, the manner in which the
  ## size is obtained is a bit simpler because the DSA only returns numeric columns.
  cv.res <- NA
  if (cross.validate.model.selected) {
    cv.call <- call
    cv.call$maxorderint <- cv.call$maxsumofpow <- 1
    cv.call$formula <- modelf
    cv.call$maxsize <- size <- length(labels(terms(modelf, data = data)))
    cv.call$cross.validate.model.selected <- FALSE
    cv.call$expand.forced.term.factors <- FALSE

    if (size > 0) {
      cv.res <- eval(cv.call, parent.frame())
      cv.res <- cv.res$average.CVrisks[1, size + 1]
    }
    else {
      warning(paste("Could not compute average cross-validated risk for model selected",
                    "Cannot perform cross-validation on intercept only model."))
    }
  }
  output <- list("average.CVrisks" = res[[4]], "min.risk" = c(gCONST[9], gCONST[8]+1),
                 "model.selected" = modelf, "family" = family$family,
                 "coefficients" = coefficients[1:(gCONST[8] + 1)],
                 "userseed" = userseed, "vsplits" = vsplits, "weights" = weights,
                 "computing.time" = total.time, "data" = data,
                 "formula" = formula, "call" = call,
                 "average.CVrisk.model.selected" = cv.res,
                 "id" = id)
  
  class(output) <- "DSA"


  ## reset some things on the way out.
  if (!missing(silent)) 
    setDSAMessageLevel(old.message.level)
  
  return(output)
}

setDSAMessageLevel <- function(newlevel) {
  storage.mode(newlevel) <- "integer"
  res <- .C("DSA_PACK_setMessagelevel", newlevel)[[1]]
}

getDSAMessageLevel <- function() {
  .C("DSA_PACK_getMessageLevel", integer(1))[[1]]
}

getDSASubversionInfo <- function() {
  svn.file <- paste(system.file(package = "DSA"), "/", "svn.info", sep = "")
  if (!file.exists(svn.file))
    stop("Package built without subversion information.")
  else
    xx <- readLines(svn.file)

  print.package.message(xx)
  
}

print.package.message <- function(svn.info = NULL) {
  if (!is.null(svn.info))
    svn.info <- strsplit(svn.info, ": ")

  cat("\n DSA version: ")
  cat(packageDescription('DSA')$Version)
  cat("\n Subversion Repository Version: ")
  if (!is.null(svn.info)) {
    cat(strsplit(svn.info[[1]][9], " ")[[1]][1])
    cat("\n Last changed date: ")
    cat(svn.info[[1]][10])
  }
  else {
    cat("Development Version")
  }
  cat("\n\n")
}

hasDSASubversionInfo <- function() {
  svn.file <- paste(system.file(package = "DSA"), "/", "svn.info", sep = "")
  if (!file.exists(svn.file))
    FALSE
  else
    TRUE
}

####################################################################################
### ################################ S3 Methods ################################ ###
####################################################################################

print.DSA <- function(x, ...) {
  cat("\nModel Selected:\n\t")
  print(x$model.selected)
  cat("\nModel Coefficients:\n\t", x$coefficients)
  cat("\nTime:\n\t", x$computing.time,attributes(x$computing.time)[[2]])
  cat("\n\n")
}

predict.DSA <- function(object, newdata = NULL, ...) {
  fit <- get.dsa.fit(object)

  if (missing(newdata)) {
    predict(fit, ...)
  }
  else {
    predict(fit, newdata, ...)
  }
}

plot.DSA <- function(x, plot.compare = FALSE, ...) {
  compare.plot <- function(dsa.res, ...) {
    plot(dsa.res$average.CVrisks[1,], xlab = "Model Size",
         ylab = "Average CV Risks", main = paste("Average Risks"),
         pch = 1)
    
    for (i in 2:nrow(dsa.res$average.CVrisks)) {
      points(dsa.res$average.CVrisks[i,], pch = i)
    }
    avg.risks <- dsa.res$average.CVrisks[which(is.finite(dsa.res$average.CVrisks))]
    avg.risks <- range(avg.risks)
    avg.risks <- avg.risks[2] - (avg.risks[2] - avg.risks[1])*.80 
    legend(ncol(dsa.res$average.CVrisks)*.6, avg.risks,
           pch = 1:nrow(dsa.res$average.CVrisks),
           legend = paste("interaction order", 1:nrow(dsa.res$average.CVrisks)))
  }
  
  if (plot.compare)
    compare.plot(x, ...)
  else {
    orig.mfrow <- par()$mfrow
    par(mfrow=n2mfrow(nrow(x$average.CVrisks)))
    for (i in 1:nrow(x$average.CVrisks)) {
      plot(x$average.CVrisks[i,], xlab = "Model Size",
           ylab = "Average CV Risks", main = paste("Interaction Order:",i))
    }
    par(mfrow=orig.mfrow)
  }
}

summary.DSA <- function(object, ...) {
  fit <- get.dsa.fit(object)
  coefficients <- as.matrix(coefficients(fit))
  colnames(coefficients) <- "Estimates"

  mm <- get.model.matrices(object$formula, data = object$data)
  df <- as.data.frame(cbind(mm$Y, mm$X))
  
  n.obs <- nrow(df)
  n.obs.nonna <- nrow(model.matrix(fit$terms, df))
  
  n.id <- length(unique(object$id))
  nid.mf <- model.frame(object$model.selected, data = df, na.action = na.pass)
  nid.mt <- attr(nid.mf, "terms")
  nid.mm <- model.matrix(nid.mt, nid.mf)[, -1]
  nid.mr <- model.response(nid.mf)
  mat <- cbind(id = object$id, nid.mr, nid.mm)

  n.id.nonna <- 0
  for (mid in unique(object$id)) {
    if (!all(is.na(rowSums(mat[mat[,1] == mid, , drop = FALSE]))))
      n.id.nonna <- n.id.nonna + 1
  }

  cvrisk <- object$average.CVrisks[object$min.risk[1], object$min.risk[2]]
  io.and.size <- list("model.size.selected" = object$min.risk[2],
                      "interaction.order.selected" = object$min.risk[1])

  summary.DSA <- list(model.selected = object$model.selected,
                      family = object$family,
                      n.obs = n.obs,
                      n.obs.nonna = n.obs.nonna,
                      best.average.CVrisk = cvrisk,
                      coefficients = coefficients,
                      io.and.size = io.and.size,
                      average.CVrisk.model.selected = object$average.CVrisk.model.selected,
                      n.id = n.id, n.id.nonna = n.id.nonna)
                      
  class(summary.DSA) <- "summary.DSA"
  return(summary.DSA)
}

print.summary.DSA <- function(x, ...) {
  cat("\nModel:\n\t")
  print(x$model)
  cat("\n")
  print(x$coefficients)
  cat("\nFamily: ")
  cat(x$family)
  cat("\nN: ")
  cat(x$n.obs)
  cat("\nN NON-NA: ")
  cat(x$n.obs.nonna)
  cat("\nN ID: ")
  cat(x$n.id)
  cat("\nN ID NON-NA: ")
  cat(x$n.id.nonna)  
  cat(paste("\nMinimum Average Cross-Validated Risk for Model Size", x$io.and.size$model.size.selected,
            "and interaction order", x$io.and.size$interaction.order.selected), ": ")
  cat(x$best.average.CVrisk)
  cat("\nAverage Cross-Validated risk for model selected: ")
  cat(x$average.CVrisk.model.selected)
  cat("\n\n")
}

coefficients.DSA <- function(object, ...) {
  return(object$coefficients)
}

residuals.DSA <- function(object, ...) {
  fit <- get.dsa.fit(object, ...)
  return(residuals(fit, ...))
}

#######################################################################################
### ############################### non-public methods ############################ ###
#######################################################################################

get.dsa.fit <- function(dsa.res) {
  model.mats <- get.model.matrices(dsa.res$formula, data = dsa.res$data)
  df <- as.data.frame(cbind(model.mats$Y,model.mats$X))
  wts <- as.numeric(dsa.res$weights[nrow(dsa.res$weights),])
  if (length(wts) == 0)
    wts <- NULL
  model <- dsa.res$model.selected
  environment(model) <- environment()
  fit <- glm(model, family = dsa.res$family, data = df, weights = wts)

  return(fit)
}

get.contrast.list <- function(formula, data) {
  mt <- terms(formula, data = data)
  variables <- attr(mt, "variables")
  response <- attr(mt, "response")
  cols <- eval(variables, data)[-response]
  var.names <- as.list(variables)[c(-1,-2)]
  
  lst <- lapply(cols, function(x) {
    if (is.factor(x)) {
      xx <- diag(1, nrow = length(levels(x)))
      colnames(xx) <- rownames(xx) <- as.character(levels(x))
      xx
    }
    else {
      NULL
    }
  })
  names(lst) <- var.names

  ##
  ## remove entries which are not null, no list if
  ## if there are no factors. (there must be a better way)
  ##
  res <- list()
  na <- c()
  if (length(lst) > 0) {
    for (i in 1:length(lst)) {
      if (!is.null(lst[[i]])) {
        res <- c(res, list(lst[[i]]))
        na <- c(na, (as.character(var.names[[i]])))
      }
    }
    names(res) <- na
  }
  if (length(res) <= 0)
    return(NULL)
  else
    return(res)
}


get.model.matrices <- function(formula, data) {
  complete.formula <- as.formula(paste(deparse(as.list(formula)[[2]]),
                                       " ~ .", sep = ""))
  complete.mf <- model.frame(complete.formula, data = data,
                             na.action = na.pass)
  mt <- attr(complete.mf, "terms")
  
  ## transpose and drop the intercept column. 
  X <- model.matrix(mt, complete.mf, contrasts.arg =
                    get.contrast.list(complete.formula, data))[,-1]

  ## preserve the fact that Ylearn is a matrix, 
  if (is.null(dim(Y <- model.response(complete.mf)))) {
    Y <- complete.mf[,1,drop = FALSE]
  }
  Y <- as.matrix(Y)
  
  return(list(X = X, Y = Y))
}

get.final.model <- function(modelselected, namesxmat, namesymat)
{
  if(sum(modelselected)==0)res <- paste(namesymat," ~ 1",sep="")
  else {
    res <- apply(modelselected,1,function(x,y) {
      vars <- y[as.logical(x)]
      powers <- x[x!=0]
      termcomponents <- apply(cbind(vars,powers),1,function(x) {
        return(paste(x,collapse="^"))
      })
      term <- ""
      if(length(termcomponents)>0)
        term <- paste(" I(", paste(termcomponents, collapse="*") ,")",sep="")
      return(term)
    }, y=namesxmat)
    if (length(namesymat) == 1) {
      res <- paste(namesymat," ~",paste(res[res!=""],collapse=" +"),sep="")
    }
    else {
      res <- paste("cbind(", namesymat[1], ",", namesymat[2], ")",
                   paste(" ~", paste(res[res!=""],collapse=" +"),sep=""))
    }
  }
  
  ff <- tryCatch(as.formula(res), simpleError = function(...) {
    warning("Unable to parse resulting formula, Potential problem with factor/level conversion.")
    return(res)
  })
  return(ff)
}

get.forced.terms <- function(formula, data, Xlearn, expand.factors) {
  forcedterms <- NULL

  if (expand.factors){
    mm <- model.matrix(terms(formula, data = data), data = data)[,-1, drop=FALSE]
  }
  else {
    df <- get.model.matrices(formula, data)
    df <- as.data.frame(cbind(df$Y, df$X))
    mm <- model.matrix(terms(formula, data = df), data = df)[,-1]
  }
                             
  Fterms <- colnames(mm)
  nFterms <- length(Fterms)
  
  if (nFterms != 0) {
    forcedterms <- matrix(0, nrow = nFterms, ncol = nrow(Xlearn))
    
    for(i in 1:nFterms) {
      ftermi <- unlist(strsplit(Fterms[i], split="*", fixed=T))
      ftermi <- unlist(apply(cbind(ftermi), 1, function(x)return(strsplit(x,split="I(",fixed=T))))
      ftermi <- unlist(apply(cbind(ftermi), 1, function(x)return(strsplit(x,split=")",fixed=T))))
      ftermi <- unlist(apply(cbind(ftermi), 1, function(x)return(strsplit(x,split=" ",fixed=T))))
      ftermi <- ftermi[ftermi != ""]
      nvari <- length(ftermi)
      powi <- rep(1,nvari)

      for(j in 1:nvari) {
        fvari <- unlist(strsplit(ftermi[j],split="^",fixed=T))
        if(length(fvari)==2) {
          powi[j] <- as.numeric(fvari[2])
          ftermi[j] <- fvari[1]
        }

        if(all(ftermi[j] != rownames(Xlearn)))
          stop("\nWARNING: Incorrect formula for forcedterms.\n")
        
        forcedterms[i, ftermi[j] == rownames(Xlearn)] <- powi[j]
      }
    }
  }
  return(forcedterms)
}


